 /*
 * odieGenericDevice.c
 *
 *	Channel drivers for a loopback communications device
 *
 * Copyright (c) YYYY ????
 *
 * See the file "????.license" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "tcl.h"
#include "tclMac.h"
#include "tclErrno.h"

/*
 * Platform Specific Includes 
 */

/*
 * Odie devices include
 */
#include "odieDevices.h"

int GNDN_Init _ANSI_ARGS_((Tcl_Interp *interp));
 
static int GNDNGetHandle _ANSI_ARGS_((ClientData instanceData,
	int direction, ClientData *handlePtr));
static int GNDNBlockMode _ANSI_ARGS_((ClientData instanceData,
	int mode));
static int GNDNInput _ANSI_ARGS_((ClientData instanceData,
	char *buf, int toRead, int *errorCode));
static int GNDNOutput _ANSI_ARGS_((ClientData instanceData,
	char *buf, int toWrite, int *errorCode));
static int  GNDNClose _ANSI_ARGS_((ClientData instanceData,
	Tcl_Interp *interp));

static int GNDNGetOptionProc  _ANSI_ARGS_((ClientData instanceData,
	Tcl_Interp *interp,
	char *optionName,Tcl_DString *dsPtr));
static int  GNDNSetOptionProc  _ANSI_ARGS_((ClientData instanceData,
	Tcl_Interp *interp,
	char *optionName,char *newVal));

static Tcl_Channel GNDNOpen _ANSI_ARGS_((
    Tcl_Interp *interp,
    ClientData devID,
    int *errorCodePtr));

static void GNDNReset _ANSI_ARGS_((ClientData devID));
static int GNDNReady _ANSI_ARGS_((PortState *infoPtr));
static void GNDNWatch _ANSI_ARGS_((ClientData instanceData,int mask));

static int TclMacDevErrorToPosixError (OSErr err);

/*
 * Internal Structures
 *
 * Any device configureation information should be stored in here
 */

/* Enumerated list of devices */
enum {
    kLoopback1 = 0,
    kLoopback2,
    kLoopback3,
    kLoopback4,
    kLoopback5
} portList;

typedef struct GNDNPort {
	unsigned char 	*buffer;	/* Storage place for transmission cond */
	int				destport;
} GNDNPort;

/*
 * This variable describes the channel type structure for GNDN based IO.
 */

static Tcl_ChannelType GNDNChannelType = {
    "gndn",					/* Type name. */
    GNDNBlockMode,		/* Set blocking or
                             * non-blocking mode.*/
    GNDNClose,			/* Close proc. */
    GNDNInput,			/* Input proc. */
    GNDNOutput,		/* Output proc. */
    NULL,					/* Seek proc. (none) */
    GNDNSetOptionProc,	/* Set option proc. */
    GNDNGetOptionProc,	/* Get option proc. */
    GNDNWatch,			/* Initialize notifier. */
    GNDNGetHandle		/* Get OS handles out of channel. */
};

static PortClass GNDNPortClass = {
	"gndn",
	GNDNOpen,
	GNDNReset,
	GNDNReady
};
	
/*
 *	Used Internally during the opening sequence to
 *	map out Tcl Port ID numbers to platform specific names
 *
 *	We have a special case for powerbook internal modems,
 *	which is a port 5, but will not map transparently over
 *	port 0 (modem) We really don't need to confuse users
 *	any more than we honestly have to.
 */

static GNDNPort GNDNPortList[] = {
	{"\p.AIn","\p.AOut"},
	{"\p.BIn","\p.BOut"},
	{"\p.CIn","\p.COut"},
	{"\p.DIn","\p.DOut"},
	{"\p.EIn","\p.EOut"},
	{NULL,NULL}
};

extern int errno;

/*
 *----------------------------------------------------------------------
 *
 * GNDN_Init --
 *
 *      Add GNDN ports to device manager
 *
 * Results:
 *	1 if readable or writable, 0 if not
 *
 * Side effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
 
int GNDN_Init(interp)
Tcl_Interp *interp;
{
	/* Backwards since AddDeviceName performs a head insert */
	Tcl_AddDeviceName("loopback.4",	"Loopback Port 4",	(ClientData) 3,&GNDNPortClass);
	Tcl_AddDeviceName("loopback.3",	"Loopback Port 3",	(ClientData) 2,&GNDNPortClass);
	Tcl_AddDeviceName("loopback.2",	"Loopback Port 2",	(ClientData) 1,&GNDNPortClass);
	Tcl_AddDeviceName("loopback.1",	"Loopback Port 1",	(ClientData) 0,&GNDNPortClass);

	Tcl_AddDeviceName("gndn4","Goes Nowhere, Does Nothing",	(ClientData) 3,&GNDNPortClass);				
	Tcl_AddDeviceName("gndn3","Goes Nowhere, Does Nothing",	(ClientData) 2,&GNDNPortClass);
	Tcl_AddDeviceName("gndn2","Goes Nowhere, Does Nothing",		(ClientData) 1,&GNDNPortClass);
	Tcl_AddDeviceName("gndn1","Goes Nowhere, Does Nothing",		(ClientData) 0,&GNDNPortClass);
	
	return TCL_OK;
}
/*
 *----------------------------------------------------------------------
 *
 * GNDNReady --
 *
 *      Tests whether the given GNDN port is readable or writable
 *      depending on the value of watchMask.
 *
 * Results:
 *	1 if readable or writable, 0 if not
 *
 * Side effects:
 *	None
 *
 *----------------------------------------------------------------------
 */

static int
GNDNReady(
    PortState *infoPtr)
{
    long length;
    OSErr err;

    /*
     * For now, we are always writable.
     */
    
    if (infoPtr->watchMask & TCL_WRITABLE) {
	return 1;
    }
    
    err = SerGetBuf(infoPtr->inputRef, &length);
    if (err != noErr) {
	return 0;
    }
    if ((infoPtr->watchMask & TCL_READABLE) && (length > 0)) {
	return 1;
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * OpenGNDNChannel --
 *
 *	Opens a Macintosh driver and creates a Tcl channel to control it.
 *
 * Results:
 *	A Tcl channel.
 *
 * Side effects:
 *	Will open a Macintosh device.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
GNDNOpen(
    Tcl_Interp *interp,
    ClientData devID,			/* Name of file to open. */
    int *errorCodePtr)			/* Where to store error code. */
{
    int channelPermissions;
    Tcl_Channel chan;
    OSErr err;
    short inputRef=0,outputRef=0;
    PortState *GNDNState;
    char channelName[64];
    int port = (int)devID;
    
    /* 
     * Detect a bad port address
     */
     if(port > kInternalModem || port < 0) {
    	*errorCodePtr = errno = EFAULT;
	Tcl_SetErrno(errno);
	return NULL;
    }
    
    /*
     * Detect an already opened port, and return it
     */
    
    sprintf(channelName, "gndn%d", (int) port);
    chan = Tcl_GetChannel(interp, channelName, &channelPermissions);
    if (chan != NULL) {
	return chan;
    } else {
	Tcl_SetStringObj(Tcl_GetObjResult(interp), NULL, 0);
    }
    
    /*
     * GNDN Channels Will always be Read/Write
     */

    channelPermissions = (TCL_READABLE | TCL_WRITABLE);

    GNDNState = (PortState *) ckalloc((unsigned) sizeof(PortState));
    chan = Tcl_CreateChannel(&GNDNChannelType, channelName,
	    (ClientData) GNDNState, channelPermissions);
	    
    if (chan == (Tcl_Channel) NULL) {
		*errorCodePtr = errno = EFAULT;
		Tcl_SetErrno(errno);
		CloseGNDNDrivers(inputRef, outputRef);
		ckfree((char *) GNDNState);
	        return NULL;
    }

    GNDNState->devChan = chan;
    GNDNState->inputRef = 0;
    GNDNState->outputRef = 0;
    GNDNState->pending = false;
    GNDNState->watchMask = 0;
    GNDNState->validMask = channelPermissions;
    GNDNState->blocking = false;

    GNDNState->devID	= devID;
    GNDNState->portClass = &GNDNPortClass;

    return chan;

    Error:
    

    Tcl_SetErrno(errno);
    
    /* 
     * Return an error condition
     */
     
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * GNDNReset --
 *
 *      Resets the input port
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Closes the drivers on the given port (opening them first if
 *      they are not already open.
 *
 *----------------------------------------------------------------------
 */

void GNDNReset (devID)
ClientData devID;
{
    short inputRef, outputRef;
	int port = (int)devID;
	return;
}

/*
 *----------------------------------------------------------------------
 *
 * GNDNBlockMode --
 *
 *	Set blocking or non-blocking mode on a GNDN channel.
 *      (Since this is done on the software level, never really errors out)
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

static int
GNDNBlockMode(
    ClientData instanceData,	/* Unused. */
    int mode)			/* The mode to set. */
{
    /*
     * Seems that yes means no, 1 - nonblocking 0 - blocking
     */
    
    PortState *GNDNState = (PortState *) instanceData;
    GNDNState->blocking = !mode;
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * DeviceWatch --
 *
 *	Initialize the notifier to watch handles from this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
GNDNWatch(
    ClientData instanceData,		/* The file state. */
    int mask)				/* Events of interest; an OR-ed
                                         * combination of TCL_READABLE,
                                         * TCL_WRITABLE and TCL_EXCEPTION. */
{
    PortState *infoPtr = (PortState *) instanceData;
    int oldMask = infoPtr->watchMask;


    infoPtr->watchMask = mask  & infoPtr->validMask;
    if (infoPtr->watchMask) {
	if (!oldMask) {
	    /*
	     * Add to notification manager
	     */
	    
	    DeviceInsert(infoPtr);
	}
    } else {
	if (oldMask) {
	    /*
	     * Remove the GNDN from the list of watched GNDNs.
	     */
	    DeviceRemove(infoPtr);
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * GNDNClose --
 *
 *	Closes the IO channel.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the physical channel
 *
 *----------------------------------------------------------------------
 */

static int
GNDNClose(
    ClientData instanceData,	/* Unused. */
    Tcl_Interp *interp)		/* Unused. */
{
    PortState *GNDNState = (PortState *) instanceData;
    int errorCode = 0;
    OSErr err;
    
    GNDNWatch(instanceData,0);

    ckfree((char *) GNDNState);
    Tcl_SetErrno(errorCode);
    return errorCode;
}
/*
 *----------------------------------------------------------------------
 *
 * GNDNInput --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

int
GNDNInput(
    ClientData instanceData,	/* Unused. */
    char *buffer,	        /* Where to store data read. */
    int bufSize,		/* How much space is available
				 * in the buffer? */
    int *errorCodePtr)		/* Where to store error code. */
{
    PortState *GNDNState = (PortState *) instanceData;
	return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * GNDNOutput--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

static int
GNDNOutput(
    ClientData instanceData,		/* Unused. */
    char *buffer,			/* The data buffer. */
    int toWrite,			/* How many bytes to write? */
    int *errorCodePtr)			/* Where to store error code. */
{
    PortState *GNDNState = (PortState *) instanceData;
    return toWrite;
}


/*
 *----------------------------------------------------------------------
 *
 * GNDNGetHandle --
 *
 *	Called from Tcl_GetChannelFile to retrieve OS handles from inside
 *	a file based channel.
 *
 * Results:
 *	The appropriate handle or NULL if not present.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
GNDNGetHandle(
    ClientData instanceData,		/* The file state. */
    int direction,			/* Which handle to retrieve? */
    ClientData *handlePtr)
{
    if (direction == TCL_READABLE) {
	*handlePtr = (ClientData) ((PortState*)instanceData)->inputRef;
	return TCL_OK;
    }
    if (direction == TCL_WRITABLE) {
	*handlePtr = (ClientData)
	    ((PortState*)instanceData)->outputRef;
	return TCL_OK;
    }
    return TCL_ERROR;
}
/*
 *----------------------------------------------------------------------
 *
 * GNDNGetOptionProc --
 *
 *	Computes an option value for a GNDN based channel, or a
 *	list of all options and their values.
 *
 *	Note: This code is based on code contributed by John Haxby.
 *
 * Results:
 *	A standard Tcl result. The value of the specified option or a
 *	list of all options and	their values is returned in the
 *	supplied DString.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
GNDNSetOptionProc(
    ClientData instanceData, 		/* Socket state. */
    Tcl_Interp *interp,                 /* For error reporting - can be NULL.*/
    char *optionName, 			/* Name of the option to
                                         * retrieve the value for, or
                                         * NULL to get all options and
                                         * their values. */
    char *newVal)			/* Where to store the computed
                                         * value; initialized by caller. */
{
    PortState *GNDNState = (PortState *) instanceData;

    return TCL_OK;
}
/*
 *----------------------------------------------------------------------
 *
 * GNDNGetOptionProc --
 *
 *	Computes an option value for a GNDN based channel, or a
 *	list of all options and their values.
 *
 *	Note: This code is based on code contributed by John Haxby.
 *
 * Results:
 *	A standard Tcl result. The value of the specified option or a
 *	list of all options and	their values is returned in the
 *	supplied DString.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
GNDNGetOptionProc(
    ClientData instanceData, 		/* Socket state. */
    Tcl_Interp *interp,                 /* For error reporting - can be NULL.*/
    char *optionName, 			/* Name of the option to
                                         * retrieve the value for, or
                                         * NULL to get all options and
                                         * their values. */
    Tcl_DString *dsPtr)			/* Where to store the computed
                                         * value; initialized by caller. */
{
    PortState *GNDNState = (PortState *) instanceData;
    return TCL_OK;

}


