/*
 * odieLoopbackDevice.c
 *
 *	Creates a set of loopback ports
 *
 * Copyright (c) 1999 Woods Design Services and Sean Woods
 *
 * See the file "odie.license" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *  Questions, Bugs reports, and Creative direction can for now be sent to
 *  Sean Woods, yoda@drexel.edu
 *
 */

#include "tcl.h"

/*
 * Tk renames Status
 * and the compiler yells at you
 */
#undef Status

#include "odieDevices.h"

int Loopback_Init _ANSI_ARGS_((Tcl_Interp *interp));
 
static int LoopbackGetHandle _ANSI_ARGS_((ClientData instanceData,
	int direction, ClientData *handlePtr));
static int LoopbackBlockMode _ANSI_ARGS_((ClientData instanceData,
	int mode));
static int LoopbackInput _ANSI_ARGS_((ClientData instanceData,
	char *buf, int toRead, int *errorCode));
static int LoopbackOutput _ANSI_ARGS_((ClientData instanceData,
	char *buf, int toWrite, int *errorCode));
static int  LoopbackClose _ANSI_ARGS_((ClientData instanceData,
	Tcl_Interp *interp));

static int LoopbackGetOptionProc  _ANSI_ARGS_((ClientData instanceData,
	Tcl_Interp *interp,
	char *optionName,Tcl_DString *dsPtr));
static int  LoopbackSetOptionProc  _ANSI_ARGS_((ClientData instanceData,
	Tcl_Interp *interp,
	char *optionName,char *newVal));

static Tcl_Channel LoopbackOpen _ANSI_ARGS_((
    Tcl_Interp *interp,
    ClientData devID,
    int *errorCodePtr));

static void LoopbackReset _ANSI_ARGS_((ClientData devID));
static int LoopbackReady _ANSI_ARGS_((PortState *infoPtr));
static void LoopbackWatch _ANSI_ARGS_((ClientData instanceData,int mask));

static int TclMacDevErrorToPosixError (OSErr err);

/*
 * This set is from the Loopback Driver Apocrypha by Apple DTS
 */

static OSErr 		CloseLoopbackDrivers(SInt16 inRefNum, SInt16 outRefNum);
static Boolean 		DriverIsOpen(ConstStr255Param driverName);
static OSErr 		OpenOneLoopbackDriver(ConstStr255Param driverName,
	short *refNum);
static Tcl_Channel	OpenLoopbackChannel _ANSI_ARGS_((int port, 
	int *errorCodePtr));
static OSErr 		OpenLoopbackDrivers _ANSI_ARGS_((ConstStr255Param inName,
	ConstStr255Param outName,
	short *inRefNum, short *outRefNum));
static Boolean 		LoopbackArbitrationExists(void);


/*
 * Internal Structures
 *
 * Any device configureation information should be stored in here
 */

enum {
    loopback1 = 0,
    kPrinter,
    kCom1,
    kCom2,
    kInternalModem
} portList;

typedef struct LoopbackPort {
	unsigned char *inref;
	unsigned char *outref;
    unsigned int serConfig;	/* Port Configuration */
    SerShk serHShake; 		/* Port Handshaking */    
} LoopbackPort;

/*
 * This variable describes the channel type structure for loopback based IO.
 */

static Tcl_ChannelType loopbackChannelType = {
    "loopback",					/* Type name. */
    LoopbackBlockMode,		/* Set blocking or
                             * non-blocking mode.*/
    LoopbackClose,			/* Close proc. */
    LoopbackInput,			/* Input proc. */
    LoopbackOutput,		/* Output proc. */
    NULL,					/* Seek proc. (none) */
    LoopbackSetOptionProc,	/* Set option proc. */
    LoopbackGetOptionProc,	/* Get option proc. */
    LoopbackWatch,			/* Initialize notifier. */
    LoopbackGetHandle		/* Get OS handles out of channel. */
};

static PortClass loopbackPortClass = {
	"loopback",
	LoopbackOpen,
	LoopbackReset,
	LoopbackReady
};
	
/*
 *	Used Internally during the opening sequence to
 *	map out Tcl Port ID numbers to platform specific names
 *
 *	We have a special case for powerbook internal modems,
 *	which is a port 5, but will not map transparently over
 *	port 0 (modem) We really don't need to confuse users
 *	any more than we honestly have to.
 */

static LoopbackPort loopbackPortList[] = {
	{"\p.AIn","\p.AOut",0,0},
	{"\p.BIn","\p.BOut",0,0},
	{"\p.CIn","\p.COut",0,0},
	{"\p.DIn","\p.DOut",0,0},
	{"\p.InternalModemIn","\p.InternalModemOut",0,0},
	{NULL,NULL,NULL,NULL}
};

/*
 * Tell if we have a loopback port arbitrator for working around problems
 * with ARA
 */

enum {
    gestaltLoopbackPortArbitratorAttr = 'arb ',
    gestaltLoopbackPortArbitratorExists = 0
};

extern int errno;

/*
 *----------------------------------------------------------------------
 *
 * LoopbackInit --
 *
 *      Add loopback ports to device manager
 *
 * Results:
 *	1 if readable or writable, 0 if not
 *
 * Side effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
 
int Loopback_Init(interp)
Tcl_Interp *interp;
{
	/* Backwards since AddDeviceName performs a head insert */
	Tcl_AddDeviceName("loopback.4",	"Loopback Port 4",	(ClientData) 3,&loopbackPortClass);
	Tcl_AddDeviceName("loopback.3",	"Loopback Port 3",	(ClientData) 2,&loopbackPortClass);
	Tcl_AddDeviceName("loopback.2",	"Loopback Port 2",	(ClientData) 1,&loopbackPortClass);
	Tcl_AddDeviceName("loopback.1",	"Loopback Port 1",	(ClientData) 0,&loopbackPortClass);

	Tcl_AddDeviceName("pbmodem",	"Powerbook Internal Modem",	(ClientData) 4,&loopbackPortClass);
	Tcl_AddDeviceName("com2",	"RS-232 Port 2",	(ClientData) 3,&loopbackPortClass);				
	Tcl_AddDeviceName("com1",	"RS-232 Port 1",	(ClientData) 2,&loopbackPortClass);
	Tcl_AddDeviceName("printer","Printer Port",		(ClientData) 1,&loopbackPortClass);
	Tcl_AddDeviceName("modem",	"Modem Port",		(ClientData) 0,&loopbackPortClass);
	
	return TCL_OK;
}
/*
 *----------------------------------------------------------------------
 *
 * LoopbackReady --
 *
 *      Tests whether the given loopback port is readable or writable
 *      depending on the value of watchMask.
 *
 * Results:
 *	1 if readable or writable, 0 if not
 *
 * Side effects:
 *	None
 *
 *----------------------------------------------------------------------
 */

static int
LoopbackReady(
    PortState *infoPtr)
{
    long length;
    OSErr err;

    /*
     * For now, we are always writable.
     */
    
    if (infoPtr->watchMask & TCL_WRITABLE) {
	return 1;
    }
    
    err = SerGetBuf(infoPtr->inputRef, &length);
    if (err != noErr) {
	return 0;
    }
    if ((infoPtr->watchMask & TCL_READABLE) && (length > 0)) {
	return 1;
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * OpenLoopbackChannel --
 *
 *	Opens a Macintosh driver and creates a Tcl channel to control it.
 *
 * Results:
 *	A Tcl channel.
 *
 * Side effects:
 *	Will open a Macintosh device.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
LoopbackOpen(
    Tcl_Interp *interp,
    ClientData devID,			/* Name of file to open. */
    int *errorCodePtr)			/* Where to store error code. */
{
    int channelPermissions;
    Tcl_Channel chan;
    OSErr err;
    short inputRef=0,outputRef=0;
    PortState *loopbackState;
    char channelName[64];
    int port = (int)devID;
    
    
    /*
     * Detect an already opened port, and return it
     */
    
    sprintf(channelName, "loopback%d", (int) port);
    chan = Tcl_GetChannel(interp, channelName, &channelPermissions);
    if (chan != NULL) {
	return chan;
    } else {
	Tcl_SetStringObj(Tcl_GetObjResult(interp), NULL, 0);
    }
    
    /*
     * Loopback Channels Will always be Read/Write
     */

    channelPermissions = (TCL_READABLE | TCL_WRITABLE);

    loopbackState = (PortState *) ckalloc((unsigned) sizeof(PortState));
    chan = Tcl_CreateChannel(&loopbackChannelType, channelName,
	    (ClientData) loopbackState, channelPermissions);
    if (chan == (Tcl_Channel) NULL) {
	*errorCodePtr = errno = EFAULT;
	Tcl_SetErrno(errno);

	ckfree((char *) loopbackState);
        return NULL;
    }
    
 

    return chan;

    Error:
    
    /* 
     * Map err to POSIX style error code
     */
     
     *errorCodePtr = errno = TclMacDevErrorToPosixError(err);
 
    /*
     * Be Sure to close out the ports before we bail
     */
    
    CloseLoopbackDrivers(inputRef, outputRef);

    Tcl_SetErrno(errno);
    
    /* 
     * Return an error condition
     */
     
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * LoopbackReset --
 *
 *      Resets the input port
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Closes the drivers on the given port (opening them first if
 *      they are not already open.
 *
 *----------------------------------------------------------------------
 */

void LoopbackReset (devID)
ClientData devID;
{
    short inputRef, outputRef;
	int port = (int)devID;

}

/*
 *----------------------------------------------------------------------
 *
 * LoopbackBlockMode --
 *
 *	Set blocking or non-blocking mode on a loopback channel.
 *      (Since this is done on the software level, never really errors out)
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

static int
LoopbackBlockMode(
    ClientData instanceData,	/* Unused. */
    int mode)			/* The mode to set. */
{
    /*
     * Seems that yes means no, 1 - nonblocking 0 - blocking
     */
    
    PortState *loopbackState = (PortState *) instanceData;
    loopbackState->blocking = !mode;
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * DeviceWatch --
 *
 *	Initialize the notifier to watch handles from this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
LoopbackWatch(
    ClientData instanceData,		/* The file state. */
    int mask)				/* Events of interest; an OR-ed
                                         * combination of TCL_READABLE,
                                         * TCL_WRITABLE and TCL_EXCEPTION. */
{
    PortState *infoPtr = (PortState *) instanceData;
    int oldMask = infoPtr->watchMask;


    infoPtr->watchMask = mask  & infoPtr->validMask;
    if (infoPtr->watchMask) {
	if (!oldMask) {
	    /*
	     * Add to notification manager
	     */
	    
	    DeviceInsert(infoPtr);
	}
    } else {
	if (oldMask) {
	    /*
	     * Remove the loopback from the list of watched loopbacks.
	     */
	    DeviceRemove(infoPtr);
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * LoopbackClose --
 *
 *	Closes the IO channel.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the physical channel
 *
 *----------------------------------------------------------------------
 */

static int
LoopbackClose(
    ClientData instanceData,	/* Unused. */
    Tcl_Interp *interp)		/* Unused. */
{
    PortState *loopbackState = (PortState *) instanceData;
    int errorCode = 0;
    OSErr err;
    
    LoopbackWatch(instanceData,0);
    err = CloseLoopbackDrivers(loopbackState->inputRef, loopbackState->outputRef);

    /*
     * Map err to a POSIX standard error code
     */
    if (err != noErr) {
	errorCode = errno = TclMacDevErrorToPosixError(err);
	/*
	 * Removed because we can recover with a port reset
	 * plus failure to close the port is peanuts compared
	 * to the havok the user will experience recovering from
	 * a crash
	 *
	 * (Usually ;)
	 *
	 panic("error during port close");
	 */
    }

    ckfree((char *) loopbackState);
    Tcl_SetErrno(errorCode);
    return errorCode;
}
/*
 *----------------------------------------------------------------------
 *
 * LoopbackInput --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

int
LoopbackInput(
    ClientData instanceData,	/* Unused. */
    char *buffer,	        /* Where to store data read. */
    int bufSize,		/* How much space is available
				 * in the buffer? */
    int *errorCodePtr)		/* Where to store error code. */
{
    PortState *loopbackState = (PortState *) instanceData;
    long length;
    OSErr err;

    if(loopbackState->blocking) {
	do {
	    length = bufSize;

	    /*
	     * Wait until information is available
	     */

	    err = SerGetBuf(loopbackState->inputRef,&length);
	    Tcl_DoOneEvent(0);
	} while(length == 0);
    }
    err = SerGetBuf(loopbackState->inputRef,&length);

    if(length > 0) {
    	if(bufSize > length)
	    bufSize = length;

    	length = bufSize;
	err = FSRead(loopbackState->inputRef, &length, buffer);
	if ((err == noErr) || (err == eofErr)) {
	    return length;
	} else {
	    *errorCodePtr = errno = TclMacDevErrorToPosixError(err);
	    Tcl_SetErrno(errno);
	    return -1;
	}
    } else {
	return 0;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * LoopbackOutput--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

static int
LoopbackOutput(
    ClientData instanceData,		/* Unused. */
    char *buffer,			/* The data buffer. */
    int toWrite,			/* How many bytes to write? */
    int *errorCodePtr)			/* Where to store error code. */
{
    PortState *loopbackState = (PortState *) instanceData;
    long length = toWrite;
    SerStaRec inSerStat, outSerStat;
    OSErr err;

    *errorCodePtr = 0;
    errno = 0;
    err = SerStatus(loopbackState->outputRef, &inSerStat);
    err = SerStatus(loopbackState->inputRef, &outSerStat);
    err = FSWrite(loopbackState->outputRef, &length, buffer);
    if (err != noErr) {
	switch(err) {
	    case controlErr:
	    case statusErr:
	    case readErr:
	    case writErr:
		*errorCodePtr = errno = EIO;
		break;
	    default:
		*errorCodePtr = errno =
		    TclMacDevErrorToPosixError(err);
	}
	Tcl_SetErrno(errno);
	return -1;
    }
    return length;
}


/*
 *----------------------------------------------------------------------
 *
 * LoopbackGetHandle --
 *
 *	Called from Tcl_GetChannelFile to retrieve OS handles from inside
 *	a file based channel.
 *
 * Results:
 *	The appropriate handle or NULL if not present.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
LoopbackGetHandle(
    ClientData instanceData,		/* The file state. */
    int direction,			/* Which handle to retrieve? */
    ClientData *handlePtr)
{
    if (direction == TCL_READABLE) {
	*handlePtr = (ClientData) ((PortState*)instanceData)->inputRef;
	return TCL_OK;
    }
    if (direction == TCL_WRITABLE) {
	*handlePtr = (ClientData)
	    ((PortState*)instanceData)->outputRef;
	return TCL_OK;
    }
    return TCL_ERROR;
}
/*
 *----------------------------------------------------------------------
 *
 * LoopbackGetOptionProc --
 *
 *	Computes an option value for a loopback based channel, or a
 *	list of all options and their values.
 *
 *	Note: This code is based on code contributed by John Haxby.
 *
 * Results:
 *	A standard Tcl result. The value of the specified option or a
 *	list of all options and	their values is returned in the
 *	supplied DString.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
LoopbackSetOptionProc(
    ClientData instanceData, 		/* Socket state. */
    Tcl_Interp *interp,                 /* For error reporting - can be NULL.*/
    char *optionName, 			/* Name of the option to
                                         * retrieve the value for, or
                                         * NULL to get all options and
                                         * their values. */
    char *newVal)			/* Where to store the computed
                                         * value; initialized by caller. */
{
    PortState *loopbackState = (PortState *) instanceData;
    int port;
    unsigned int temp;
    unsigned int dataconv;
    OSErr	err;

    port = (int)loopbackState->devID;
    temp = loopbackPortList[port].serConfig;

    /*
     * Determine which options we need to modify.
     */

    if (!strcmp(optionName, "-baud")) {
	int baud;
	/*
	 * Mask Out Baud Bits
	 */
	
	temp = temp & 0xFE00;
	dataconv = atoi(newVal);
	switch (dataconv) {
	    case 150: 		baud = baud150; break;
	    case 300: 		baud = baud300; break;
	    case 600: 		baud = baud600; break;
	    case 1200:		baud = baud1200; break;
	    case 1800: 		baud = baud1800; break;
	    case 2400: 		baud = baud2400; break;
	    case 3600: 		baud = baud3600; break;
	    case 4800:		baud = baud4800; break;
	    case 7200: 		baud = baud7200; break;
	    case 9600:		baud = baud9600; break;
	    case 14400: 	baud = baud14400; break;
	    case 19200:		baud = baud19200; break;
	    case 28800: 	baud = baud28800; break;
	    case 38400: 	baud = baud38400; break;
	    case 57600: 	baud = baud57600; break;
	    default:
		if (interp) {
		    Tcl_AppendResult(interp, "bad value for -baud: ",
			    "valid settings are",
			    "150 300 600 1200 ",
			    "1800 2400 3600 4800 ",
			    "7200 9600 14400 19200 ",
			    "28800 38400 57600",
			    (char *) NULL);
		    return TCL_ERROR;
		}
	}
	temp = temp | baud;
    } else if (!strcmp(optionName, "-databits")) {
	int bits;
	temp = temp & 0xF3FF;
	dataconv = atoi(newVal);
	switch (dataconv) {
	    case 5: 		bits = data5; break;
	    case 6: 		bits = data6; break;
	    case 7: 		bits = data7; break;
	    case 8:			bits = data8; break;
	    default:
		if (interp) {
		    Tcl_AppendResult(interp, "bad value for -databits: ",
			    "valid settings are",
			    "5 6 7 8 ",
			    (char *) NULL);
		    return TCL_ERROR;
		}
	}
	temp = temp | bits;
    } else if (!strcmp(optionName, "-parity")) {
	temp = temp & 0xCFFF;
	if (!strcmp(newVal, "none")) {
	    temp = temp | noParity;
	} else if (!strcmp(newVal, "odd")) {
	    temp = temp | oddParity;
	} else if (!strcmp(newVal, "even")) {
	    temp = temp | evenParity;
	} else {
	    if (interp) {
                Tcl_AppendResult(interp, "bad value for -parity: ",
                        "valid settings are",
                        "none odd even ",
                        (char *) NULL);
                return TCL_ERROR;
	    }
	}
    } else if (!strcmp(optionName, "-stopbits")) {
	temp = temp & 0x3FFF;

	if (!strcmp(newVal, "1")) {
	    temp = temp | 0x4000;
	} else if (!strcmp(newVal, "1.5")) {
	    temp = temp | 0x8000;
	} else if (!strcmp(newVal, "2")) {
	    temp = temp | 0xC000;
	} else {
	    if (interp) {
                Tcl_AppendResult(interp, "bad value for -stopbits: ",
                        "valid settings are",
                        "1 1.5 2 ",
                        (char *) NULL);
                return TCL_ERROR;
	    }
	}
    } else {
	return Tcl_BadChannelOption(interp, optionName,
		"baud databits parity stopbits");
    }
    loopbackPortList[port].serConfig = temp;
    /*
     * Port Communication Settings/Reset
     */
    
    err = SerReset(loopbackState->inputRef,loopbackPortList[port].serConfig);
    if (err != noErr) goto Error;
    err = SerReset(loopbackState->outputRef,loopbackPortList[port].serConfig);
    if (err != noErr) goto Error;

    return TCL_OK;

    Error:
    errno = TclMacDevErrorToPosixError(err);
    Tcl_SetErrno(errno);
    return TCL_ERROR;
}
/*
 *----------------------------------------------------------------------
 *
 * LoopbackGetOptionProc --
 *
 *	Computes an option value for a loopback based channel, or a
 *	list of all options and their values.
 *
 *	Note: This code is based on code contributed by John Haxby.
 *
 * Results:
 *	A standard Tcl result. The value of the specified option or a
 *	list of all options and	their values is returned in the
 *	supplied DString.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
LoopbackGetOptionProc(
    ClientData instanceData, 		/* Socket state. */
    Tcl_Interp *interp,                 /* For error reporting - can be NULL.*/
    char *optionName, 			/* Name of the option to
                                         * retrieve the value for, or
                                         * NULL to get all options and
                                         * their values. */
    Tcl_DString *dsPtr)			/* Where to store the computed
                                         * value; initialized by caller. */
{
    PortState *loopbackState = (PortState *) instanceData;
    int doAll=0,
	doBaud=0,
	doDataBits=0,
	doParity=0,
	doStopBits=0;
    char buffer[128];
    int port;

    port = (int)loopbackState->devID;
    /*
     * Determine which options we need to do.  Do all of them
     * if optionName is NULL.
     */

    if (optionName == (char *) NULL || optionName[0] == '\0') {
        doAll = true;
    } else {
	if (!strcmp(optionName, "-baud")) {
	    doBaud = true;
	} else if (!strcmp(optionName, "-databits")) {
	    doDataBits = true;
	} else if (!strcmp(optionName, "-parity")) {
	    doParity = true;
	} else if (!strcmp(optionName, "-stopbits")) {
	    doStopBits = true;
	} else {
	    return Tcl_BadChannelOption(interp, optionName,
		    "baud databits parity stopbits");
	}
    }


    /*
     * Get the sockname for the socket.
     */

    if (doAll || doBaud) {
    	int baud;
	if (doAll) {
	    Tcl_DStringAppendElement(dsPtr, "-baud");
	}
	switch (loopbackPortList[port].serConfig & 0x1FF) {
	    case baud150: baud = 150; break;
	    case baud300: baud = 300; break;
	    case baud600: baud = 600; break;
	    case baud1200: baud = 1200; break;
	    case baud1800: baud = 1800; break;
	    case baud2400: baud = 2400; break;
	    case baud3600: baud = 3600; break;
	    case baud4800: baud = 4800; break;
	    case baud7200: baud = 7200; break;
	    case baud9600: baud = 9600; break;
	    case baud14400: baud = 14400; break;
	    case baud19200: baud = 19200; break;
	    case baud28800: baud = 28800; break;
	    case baud38400: baud = 38400; break;
	    case baud57600: baud = 57600; break;
	}
	sprintf(buffer, "%d", baud);
	Tcl_DStringAppendElement(dsPtr, buffer);
    }
    if (doAll || doDataBits) {
	if (doAll) {
	    Tcl_DStringAppendElement(dsPtr, "-databits");
	}
	switch (loopbackPortList[port].serConfig & 0xC00) {
	    case data5:		sprintf(buffer, "5"); break;
	    case data6: 	sprintf(buffer, "6"); break;
	    case data7: 	sprintf(buffer, "7"); break;
	    case data8: 	sprintf(buffer, "8"); break;
	}
	Tcl_DStringAppendElement(dsPtr, buffer);
    }
    if (doAll || doParity) {
	if (doAll) {
	    Tcl_DStringAppendElement(dsPtr, "-parity");
	}
	switch (loopbackPortList[port].serConfig & 0x3000) {
	    case noParity:		sprintf(buffer, "none"); break;
	    case oddParity: 	sprintf(buffer, "odd"); break;
	    case evenParity: 	sprintf(buffer, "even"); break;
	}
	Tcl_DStringAppendElement(dsPtr, buffer);
    }
    if (doAll || doStopBits) {
	if (doAll) {
	    Tcl_DStringAppendElement(dsPtr, "-stopbits");
	}
	switch (loopbackPortList[port].serConfig & 0xC000) {
	    case 0x4000:	sprintf(buffer, "1"); break;
	    case 0x8000: 	sprintf(buffer, "1.5"); break;
	    case 0xC000: 	sprintf(buffer, "2"); break;
	}
	Tcl_DStringAppendElement(dsPtr, buffer);
    }
    return TCL_OK;
}

