/*
 * odieYgg.c
 *
 *	Implimements ODIE Hypertree (yggdrasil)
 *
 * Copyright (c) 1999 Woods Design Services and Sean Woods
 *
 * See the file "odie.license" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Latest Changes:
 * Added semaphores to prevent the task from switching
 * in the middle of name resolution
 */

#include <tcl.h>
#include <tclInt.h>
#include "ygg.h"

/* Internal Operations */
static YggNode *NodeAdd _ANSI_ARGS_((YggNode *Node, int tree, char *name,YggNode *newnode));
static YggNode *NodeRemove _ANSI_ARGS_ ((YggNode *pNode,int tree,YggNode *thisNode));
static YggNode *NodeCreate _ANSI_ARGS_ (());

static void NodeFree _ANSI_ARGS_ ((YggNode *node));
static void NodeFreeData _ANSI_ARGS_ ((YggNode *node));
static YggNode * NodeAssociate _ANSI_ARGS_ ((char *name,YggNode *newNode));
static void NodeCopy _ANSI_ARGS_ ((YggNode *srcNode,YggNode *destNode));

static void NodeDelete _ANSI_ARGS_ ((YggNode *thisNode));

static int NodeInvalid _ANSI_ARGS_ ((YggNode *node));
static YggNode *NodeSearchTree _ANSI_ARGS_((int thisSearch,char *token,YggNode *thisNode,int *errcode));
static void NodeInsert _ANSI_ARGS_((YggNode *pNode,int tree,YggNode *thisNode));
static YggNode *NodeSort _ANSI_ARGS_ ((YggNode *headptr));
static Tcl_DString * NodeAddress _ANSI_ARGS_ ((YggNode *record));

static void BranchExpand _ANSI_ARGS_ ((YggNode *node,int recurse,Tcl_DString *dPtr));
static inline int TokenNext _ANSI_ARGS_ ((char *name,char *token,int *idx,int namelen));

/* Tcl Resolver Code */
static Tcl_Var ResolveRuntimeVarProc _ANSI_ARGS_((
    Tcl_Interp* interp, struct Tcl_ResolvedVarInfo *vinfoPtr));
static void ResolveVarDeleteProc _ANSI_ARGS_((
    struct Tcl_ResolvedVarInfo *vinfoPtr));
static int ResolveCompiledVarProc _ANSI_ARGS_((
    Tcl_Interp* interp, char* name, int length,
    Tcl_Namespace *context, Tcl_ResolvedVarInfo **rPtr));
static int ResolveVarProc _ANSI_ARGS_((
    Tcl_Interp* interp, char* name, Tcl_Namespace *context,
    int flags, Tcl_Var *rPtr));
static int ResolveCmdProc _ANSI_ARGS_((Tcl_Interp* interp,
	char* name, Tcl_Namespace *context, int flags,
	Tcl_Command *rPtr));

/*
 * The first pointer
 * of the database
 */
static YggNode 			*YggRoot = NULL;
static YggNode 			*YggTrash = NULL;

static char				scriptbuf[1024];
static char				*YggErrorString = NULL;
static char				*YggErrorInfo = NULL;
static Namespace		*globalNsPtr = NULL;
static Tcl_Interp		*INTERP;

static int				YggInterps = 0;
static int				YggNodes = 0;

static int				YggMutex = 0;
/* 
 * Common place to store names 
 * Nodes with the same references the
 * same pointer
 */

static int	nameTableLength = 0;
static char **nameTableData = NULL;

static char *YggErrors[] = {
	"",
	"Unknown Error: ",
	"Node Not Found: ",
	"Node Already Exists: ",
	"Bad Address: ",
	"Bad Token: ",
	"Bad Root Operation",
	"Invalid Record Number: ",
	"Unimplimented Command",
	"Not a Tcl Variable: ",
	"Memory Error",
	"Illegal Character in Address",
	"Last Error In Table",
	(char *)NULL
};

enum {
	yggErrNone,
	yggErrUnknown,
	yggErrNodeNotFound,
	yggErrNodeExists,
	yggErrBadAddress,
	yggErrBadToken,
	yggBadRootOp,
	yggErrRecordNumber,
	yggErrUnkCommand,
	yggErrNotTclVar,
	yggErrMemory,
	yggIllegalToken,
	yggErrLastErr
} YggErrorCode;

int Ygg_Init(interp) 
	Tcl_Interp *interp;
{
	Interp *locInterp = (Interp *)interp; 
	YggNode *irootnode;
			
	Tcl_AddInterpResolvers(
		interp, 
		"Yggdrasil", 
		NULL, 
		ResolveVarProc,
		NULL);
		
	if(YggRoot == NULL) {
		YggRoot = (YggNode *) Tcl_Alloc(sizeof(YggNode));
	 	YggRoot->name[0] = '\0';
	 	YggRoot->data = NULL;
 		YggRoot->itree = NULL;
 		YggRoot->jtree = NULL;
 		YggRoot->parent = NULL;
 	 	YggRoot->parenttype = 0;	
 		YggRoot->contents = 0;
 		YggRoot->nextPtr = NULL;
	}

	sprintf(scriptbuf,"interp%d",YggInterps++);
	irootnode = NodeAdd(YggRoot,yggItree,scriptbuf,NULL);		
		
	if(globalNsPtr == NULL) {
		globalNsPtr = locInterp->globalNsPtr;	
	}
	Tcl_CreateObjCommand(interp, "tree", Ygg_ObjCmd, NULL, NULL);	
	Tcl_CreateObjCommand(interp, "ygg", Ygg_ObjCmd, NULL, NULL);
	Tcl_CreateObjCommand(interp, "yggutil", Ygg_ObjExtraCmd, NULL, NULL);	
	return TCL_OK;
}

/*
 * Tree Command Core Functions
 */
 Ygg_ObjCmd(dummy, interp, objc, objv) 
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
	char *name;
	int namelen=0;
	YggNode *thisNode;
	Tcl_Obj *result;
	static char *subCmds[] = {
		"create",
		"set",		
		"parent",
		"parenttype",
		"children",
		"exists",
		"rm",
		"cp",
		"ln",
		"mv",
		"tail",
		(char *) NULL
	};
		
	enum {
		yCdCreate,
		yCdSet,
		yCdParent,
		yCdParentType,
		yCdChildren,
		yCdExists,
		yCdDelete,
		yCdCopy,
		yCdLink,		
		yCdMove,
		yCdTail
	}index;	
	INTERP = interp;
	result = Tcl_GetObjResult(interp);
	if (objc < 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "command ?arg?");		
		return TCL_ERROR;		
	}	

    if (Tcl_GetIndexFromObj(interp, objv[1], subCmds, "command", 0,
	    (int *) &index) != TCL_OK) 
	    return TCL_ERROR;

	Tcl_ResetResult(interp);
	switch(index) {
		case yCdCreate:
		{	
			if (objc != 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "create address");		
				return TCL_ERROR;
			}							
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			thisNode = Ygg_Record(name,namelen,1);
			return TCL_OK;
		}
		case yCdSet:
		{	
			Tcl_Obj *newobj;

			if (objc != 4) {
				Tcl_WrongNumArgs(interp, 1, objv, "set address val");		
				return TCL_ERROR;
			}				

			name = Tcl_GetStringFromObj(objv[2], &namelen);
			
			thisNode = Ygg_Record(name,namelen,1);
		  	if(thisNode == NULL) {
		  		goto error;		
			}
			newobj = Tcl_DuplicateObj(objv[3]);
			Ygg_SetObj(thisNode,newobj);
			return TCL_OK;
		}
		
		case yCdParent: {
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "record address");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			thisNode = Ygg_Record(name,namelen,0);
			if(thisNode == NULL) {
				goto error;					
			}
			if(thisNode->parent != NULL) {
				Tcl_DString *dPtr;
				dPtr = NodeAddress(thisNode->parent);
				Tcl_DStringResult(interp, dPtr);
				Tcl_DStringFree(dPtr);
				Tcl_Free((char *)dPtr);
			} else {
				Tcl_ResetResult(interp);			
			}
			return TCL_OK;		
		}
		case yCdParentType: {
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "record address");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			thisNode = Ygg_Record(name,namelen,0);
			if(thisNode == NULL) {
				goto error;				
			}
			switch(thisNode->parenttype) {
				case yggItree:
					Tcl_SetStringObj(result,
						"i", -1);
					break;
				case yggJtree:
					Tcl_SetStringObj(result,
						"j", -1);
					break;				
				default:
					Tcl_SetStringObj(result,"", -1);
					break;
			}
			return TCL_OK;		
		}
		case yCdChildren:
		{	int tree;
		
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "children address");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
  			if(name[0] != '/') {
  				Ygg_Error(yggErrBadAddress,name);
				goto error;  			
  			}

  			if(namelen == 1) {
  				thisNode = YggRoot->itree;
  			} else {
 				if(name[namelen - 1] == '/') {
					name[namelen - 1] = NULL;
					tree = yggItree;
				} else {
					tree = yggJtree;
				}
				thisNode = Ygg_Record(name,namelen,0);
				if(thisNode == NULL) {
					goto error;					
				}
				if(tree == yggItree) {
					thisNode = thisNode -> itree;
				} else {
					thisNode = thisNode -> jtree;			
				}				
			}	
			while(thisNode != NULL) {
				Tcl_AppendToObj(result, thisNode->name, -1);
				Tcl_AppendToObj(result, " ", -1);
				thisNode = thisNode -> nextPtr;		
			}
			return TCL_OK;
		}
		case yCdExists:
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "exists address");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			thisNode = Ygg_Record(name,namelen,0);
			if(thisNode == NULL) {
				Tcl_SetBooleanObj(result,0);			
			} else {
				Tcl_SetBooleanObj(result,1);
			}
			return TCL_OK;					
		case yCdDelete:
		{
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "rm address");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			thisNode = Ygg_Record(name,-1,0);
			if (thisNode == NULL) {
				Ygg_Error(yggErrNodeNotFound,name);	
				goto error;
			}
			if (thisNode == YggRoot) {
				Ygg_Error(yggBadRootOp,"");
				goto error;
			}
			NodeRemove(thisNode->parent,thisNode->parenttype,thisNode);
			NodeDelete(thisNode);
			return TCL_OK;
		}
		case yCdCopy:
		case yCdLink:		
		case yCdMove: {		
			char *src;
			char *dest;
			int srclen,destlen;
			int opcode;
			if (objc < 4) {
				Tcl_WrongNumArgs(interp, 1, objv, "op source dest ?-force? ?-recurse?");		
				return TCL_ERROR;		
			}
			src = Tcl_GetStringFromObj(objv[2],&srclen);
			dest = Tcl_GetStringFromObj(objv[3],&destlen);
			opcode = index - yCdCopy;
			
  			if(Ygg_Operation(src,srclen,dest,destlen,opcode+1) != TCL_OK) {
				goto error;
  			}
  			return TCL_OK;
  		}
		case yCdTail:
		{
			int x;
			char nchar;
			char buffer[YGGMAXTOKENLEN];
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "tail address");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			for (x = 1; x <= YGGMAXTOKENLEN, x<=namelen ; x++) {
				nchar = name[namelen-x];
				if(nchar == '.') break;
				if(nchar == '/') break;
				buffer[YGGMAXTOKENLEN - x] = nchar;
			}
			Tcl_SetStringObj(result,(buffer + YGGMAXTOKENLEN - x + 1), x-1);
			return TCL_OK;
		}
		default:
			/* If Command was not found internally,
			 * must've been a goof
			 */
			Ygg_Error(yggErrUnkCommand,"");
			goto error;
	}
error:
	Tcl_ResetResult(interp);
	Tcl_SetStringObj(result,YggErrorString, -1);
	Tcl_AppendToObj(result,YggErrorInfo,-1);		
	return TCL_ERROR;
}

/*
 * Generate an error
 */	
 
void Ygg_Error(int errCode,char *info) {
	int Len;
	if(errCode < 0 || errCode >= yggErrLastErr) {
		errCode = yggErrUnknown;
	}
	YggErrorString = YggErrors[errCode];
	Len = strlen(info);
	if(YggErrorInfo != NULL) {
		Tcl_Free(YggErrorInfo);
	}
	YggErrorInfo = Tcl_Alloc(Len);
	YggErrorInfo = strcpy(YggErrorInfo,info);
}

/*
 * Extended Tree Functions
 */
 
int Ygg_ObjExtraCmd(dummy, interp, objc, objv) 
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
	char *name;
	YggNode *thisNode;
	int namelen=0;
	Tcl_Obj *result;
	char token[YGGMAXTOKENLEN];
	static char *subCmds[] = {
		"export",
		"import",
		"nexttoken",
		"split",
		"nodes",
		"record",
		"address",		
		(char *) NULL 	
	};

	enum {		
		yCdCondense,
		yCdExpand,
		yCdNextToken,
		yCdSplit,
		yCdNodes,
		yCdRecord,
		yCdAddress,			
	}index;
	
	result = Tcl_GetObjResult(interp);
	if (objc < 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "command ?arg?");		
		return TCL_ERROR;		
	}	

    if (Tcl_GetIndexFromObj(interp, objv[1], subCmds, "command", 0,
	    (int *) &index) != TCL_OK) 
	    return TCL_ERROR;
	Tcl_ResetResult(interp);	
	switch(index) {
		case yCdRecord:
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "record address");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			thisNode = Ygg_Record(name,namelen,0);
			if(thisNode == NULL) {
				goto error;					
			}
			sprintf(scriptbuf,"/#%x",thisNode);
			Tcl_SetStringObj(result,
					scriptbuf, -1);			
			return TCL_OK;
		case yCdAddress: {
			long val;
			YggNode *thisNode;
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "address record");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			if(name[0] != '/' || name[1] != '#') {
				Ygg_Error(yggErrBadAddress,name);
				goto error;					
			}
			sscanf(name,"/#%x",&val);
			thisNode = (YggNode *)val;
			if (NodeInvalid(thisNode))  {
				Ygg_Error(yggErrRecordNumber,name);	
				goto error;		
			}

			name = Ygg_Address(thisNode);
			Tcl_SetStringObj(result,name, -1);
			return TCL_OK;		
		}
		case yCdCondense: {
			static char *flags[] = {
				"-ti",
				"-tj",
				"-ri",
				"-rj",
				(char *)NULL
			};
			enum {
				flgTi,
				flgTj,
				flgRi,
				flgRj
			} flgIdx;
			int root = (yggItree | yggJtree);
			int recurse = (yggItree | yggJtree);
			int i;
			Tcl_DString *dPtr;
						
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "consense addr ?-ti -tj -ri -rj?");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
			thisNode = Ygg_Record(name,namelen,0);
			for(i=3;i<objc;i++) {
				if(Tcl_GetIndexFromObj(interp, objv[1], flags, "options", 0,
	 			   (int *) &flgIdx) != TCL_OK) { 
	  			  return TCL_ERROR;
	  			}
	  			switch(flgIdx) {
	  				case flgTi: root ^= yggItree; break;
	  				case flgTj: root ^= yggJtree; break;
	  				case flgRi: recurse ^= yggItree; break;
	  				case flgRj: recurse ^= yggJtree; break;
	  			}
			}
		
			dPtr = (Tcl_DString *) Tcl_Alloc(sizeof(Tcl_DString));
			Tcl_DStringInit(dPtr);
			Tcl_DStringSetLength(dPtr,0);
			if((root & yggItree) != 0) {
				BranchExpand(thisNode->itree,recurse,dPtr);		
			}
			if((root & yggJtree) != 0) {
				BranchExpand(thisNode->jtree,recurse,dPtr);		
			}
			Tcl_DStringResult(interp, dPtr);
			Tcl_Free((char *)dPtr);
			return TCL_OK;
		}

		case yCdNextToken:
		{	
			int idx,reply;
			if (objc < 4) {
				Tcl_WrongNumArgs(interp, 1, objv, "nexttoken address index");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
  			if (Tcl_GetIntFromObj(interp,objv[3], &idx) != TCL_OK) {
  				return TCL_ERROR;
  			}

  			if(name[0] != '/') {
				Ygg_Error(yggErrBadAddress,name);
				goto error; 
  			}

  			reply = TokenNext(name,token,&idx,namelen);
  			if (reply == yggError) {
				goto error;
			} 	
  			Tcl_SetStringObj(result,token, -1);
  			switch (reply) {
  				case yggItree: Tcl_AppendToObj(result, " i ", -1);	break;
  				case yggJtree: Tcl_AppendToObj(result, " j ", -1); break;
  				case yggEnd:   Tcl_AppendToObj(result, " x ", -1); break;						
			}			
			return TCL_OK;
		}
		case yCdSplit:
		{	
			int idx,reply;
			if (objc < 3) {
				Tcl_WrongNumArgs(interp, 1, objv, "split address");		
				return TCL_ERROR;		
			}
			name = Tcl_GetStringFromObj(objv[2], &namelen);
  			idx = 0;

  			if(name[0] != '/') {
				Ygg_Error(yggErrBadAddress,name);
				goto error; 
  			}
			Tcl_SetStringObj(result, "{i", -1);
			while(1) {
  			reply = TokenNext(name,token,&idx,namelen);
  			if (reply == yggError) {
				goto error;
			} 
			
  			Tcl_AppendToObj(result,token, -1);
  			switch (reply) {
  				case yggItree: Tcl_AppendToObj(result, "} {i ", -1);	break;
  				case yggJtree: Tcl_AppendToObj(result, "} {j ", -1); break;
  				case yggEnd:   Tcl_AppendToObj(result, "}", -1); return TCL_OK;						
			}
			}
			return TCL_OK;
		}
		case yCdNodes:
			Tcl_SetIntObj(result,YggNodes);
			return TCL_OK;	
		default:
			/* If Command was not found internally,
			 * must've been a goof
			 */
			Ygg_Error(yggErrUnkCommand,"");
			goto error;	
	}
error:
	Tcl_ResetResult(interp);
	Tcl_SetStringObj(result,YggErrorString, -1);
	Tcl_AppendToObj(result,YggErrorInfo,-1);		
	return TCL_ERROR;
}
#ifdef PROTOTYPE

#else

/*
 * Finds a node in a tree
 */

YggNode * Ygg_Record(name,namelen,create)
	char *name;
	int namelen;
	int create;
{
	YggNode *thisNode,*nextNode;
	char token[YGGMAXTOKENLEN];
	int idx;
	int nextSearch,thisSearch;
	int found=0;
	int err = TCL_OK;
	
	if(namelen < 0) {
		namelen = strlen(name);
	}
	/*
	 * All Ygg addresses start at
	 * root
	 */
	if (name[0] != '/') {
		Ygg_Error(yggErrBadAddress,name);
		return NULL;
	}
	
	if (name[1] == '#') {
		long val;
	
		sscanf(name,"/#%x",&val);
		thisNode = (YggNode *)val;
		if (NodeInvalid(thisNode))  {
			Ygg_Error(yggErrRecordNumber,name);	
			return NULL;				
		}
		return thisNode; 
	}
	if (name[1] == NULL) {
		return YggRoot;
	}
		
	thisNode = YggRoot;
	idx = 0;	

	/*
	 * Handle Special Case
	 */

	thisSearch = yggItree;
	
	while(1) {
		/*
		 * Get the next token
		 */
		nextSearch = TokenNext(name,token,&idx,namelen);
		if(nextSearch == yggError) {
			/* Error Parsing Address */
			return NULL;
		}
		/*
		 * Scan thisSearch for a node of name 'token'
		 */
		nextNode = NodeSearchTree(thisSearch,token,thisNode,&err);
		if(err != TCL_OK) {
			/* Encountered Some Error, Return */
			return NULL;
		}

		if(nextNode == NULL) {
			break; /* Child Not Found */
		} else {
			thisNode = nextNode;
		}
		if(nextSearch == yggEnd) {
			goto done;
		}
		/*
		 * Ready for next token
		 */
		thisSearch = nextSearch;
	}

	/*
	 * Node in question was not found
	 * Do I need to create it?
	 */
	if(!create) {
		Ygg_Error(yggErrNodeNotFound,name);
		return NULL;
	}
	/*
	 * Starting from the last good parent
	 * Build a complete adress
	 */
	while(thisSearch != yggEnd) {
		thisNode = NodeAdd(thisNode,thisSearch,token,NULL);
		thisSearch = nextSearch;
		nextSearch = TokenNext(name,token,&idx,namelen);
		if(nextSearch == yggError) {
			/* Error Parsing Address */
			return NULL;
		}
	}
done:
	if (NodeInvalid(thisNode)) {	
		Ygg_Error(yggErrMemory,name);	
		return NULL;
	}
	return thisNode;
}

#endif
/*
 * Generate an address from a node
 */
 
char * Ygg_Address (record)
	YggNode *record;
{
	char *buffer;
	Tcl_DString *dPtr;
	
	dPtr = NodeAddress(record);
	buffer =  Tcl_DStringValue(dPtr);
	Tcl_DStringFree(dPtr);
	Tcl_Free((char *)dPtr);	
	return buffer;
}

/*
 * Performs a structural operation on a node
 *
 * delete,copy, move, link
 */
int Ygg_Operation(src,srclen,dest,destlen,opcode)
	char *src;
	int srclen;
	char *dest;
	int destlen;
	int opcode;
{
	YggNode *thisNode,*destNode;
	
	if(srclen < 0) {
		srclen = strlen(src);
	}
	if(destlen < 0) {
		destlen = strlen(dest);
	}
	thisNode = Ygg_Record(src,srclen,0);
	if (thisNode == NULL) {
		return TCL_ERROR;
	}
	if (thisNode == YggRoot) {
		Ygg_Error(yggBadRootOp,"");	
		return TCL_ERROR;
	}

	/* 
	 * If Destination Exists,
	 * return an error
	 */
	destNode = Ygg_Record(dest,destlen,0);
	if (destNode != NULL) {
		Ygg_Error(yggErrNodeExists,dest);
		return TCL_ERROR;
	}

	
	switch (opcode & 3) {
		case yggOpLink:
			/* 
			 * Resolver will transparently hop over
			 * to origional
			 */
			/*
			 * If we can't create destination
			 * return an error
			 */
			destNode = Ygg_Record(dest,destlen,1);
			if (destNode == NULL) {
				return TCL_ERROR;
			}			 
			destNode->contents = yggLink;
			destNode->data = (ClientData) thisNode;		
			return TCL_OK;
			
		case yggOpCopy:
		{
			/*
			 * If we can't create destination
			 * return an error
			 */
			destNode = Ygg_Record(dest,destlen,1);
			if (destNode == NULL) {
				return TCL_ERROR;
			}	
			NodeCopy(thisNode,destNode);
			return TCL_OK;
		}
		case yggOpMove:
		{	
			NodeRemove(thisNode->parent,thisNode->parenttype,thisNode);
			
			if(destlen == 0) {
				NodeDelete(thisNode);
			} else {
				destNode = Ygg_Record(dest,destlen,1);
				if (destNode == NULL) {
					return TCL_ERROR;
				}
			
				destNode->itree = thisNode->itree;
				destNode->jtree = thisNode->jtree;
				destNode->data = thisNode -> data;
				Tcl_Free((char *)thisNode);	
			}
			return TCL_OK;
		}
		default:
			Ygg_Error(yggErrUnkCommand,"");				
			return TCL_ERROR;					
	}
	return TCL_OK;
}


/*
 * TclVar Access Functions
 *
 */

/*
 * Add/Replace the contents of a node
 */
 
void Ygg_SetObj(node,objptr)
	YggNode *node;
	Tcl_Obj *objptr;
{
	Var *varPtr = NULL;
	
	NodeFreeData(node);
		
	varPtr = (Var *) Tcl_Alloc(sizeof(Var));
    varPtr->name = NULL;
 	varPtr->flags = (VAR_SCALAR | VAR_NAMESPACE_VAR | VAR_RESOLVED);
	varPtr->nsPtr = globalNsPtr;
	varPtr->refCount = 1;	/* protect from being deleted */
    varPtr->tracePtr = NULL;
	varPtr->searchPtr = NULL;			
	/*
	 *  NOTE:  Tcl reports a "dangling upvar" error for variables
	 *         with a null "hPtr" field.  Put something non-zero
	 *         in here to keep Tcl_SetVar2() happy.  The only time
	 *         this field is really used is it remove a variable
	 *         from the hash table that contains it in CleanupVar,
	 *         but since these variables are protected by their
     *         higher refCount, they will not be deleted by CleanupVar
     *         anyway.  These variables are unset and removed in
     *         ItclFreeObject().
	 *
	 *		   I Dunno. I keep loosing stuff, so I copied verbatim
	 *		   from Incr Tcl.
	 *			- Yoda
	 */        
	varPtr->hPtr = (Tcl_HashEntry*)0x1;
	varPtr->hPtr = NULL;
	

	varPtr->value.objPtr = objptr;
	
	/* Mark the data as shared */
	Tcl_IncrRefCount(varPtr->value.objPtr);	
	
	node->contents = yggTclObj;
	node->data =(ClientData)varPtr;
}

/*
 * Retrieve the object stored in a node
 */
 
Tcl_Obj * Ygg_GetObj(node)
	YggNode *node;
{
	return ((Var *)node->data)->value.objPtr;
}

/*
 * Yggdrasil equivilent of Tcl_Linkvar
 */
 
int Ygg_TreeLink(interp,name,ptr,code)
	Tcl_Interp *interp;
	char *name;
	char *ptr;
	int code;
{
	YggNode *thisNode;
	Tcl_Obj *valptr;
	int error;
	/*
	 * Find an entry in the tree, 
	 * or create new one if not found
	 */
	thisNode = Ygg_Record(name,-1,1);
	
	/*
	 * Place an empty value in the tree
	 */
	valptr = Tcl_NewObj();
	Ygg_SetObj(thisNode,valptr);
	error = Tcl_LinkVar(interp,name,(char *)ptr,code);
	return error;
}

/*
 * Internal Tree Routines
 * NO USER SERVICEABLE PARTS BELOW THIS POINT
 */
 
/*
 * Adds a new branch to a node
 * Returns the new node
 */
 
 YggNode *NodeAdd(Node,tree,name,newnode)
 	YggNode *Node;
 	int tree;
 	char *name;
  	YggNode *newnode;
 {
 	int len;

	if(newnode == NULL) {
		newnode = NodeCreate();
 	}
 	/* For now, copy
 	 * Later on I want to store all tokens in a list
 	 * To use similar ones
 	 */
 	for(len = 0; len < YGGMAXTOKENLEN;len++) {
 		newnode->name[len] = name[len];
 		if(name[len] == NULL) {
 			break;
 		}
 	}

 	newnode->parent = Node;
  	newnode->parenttype = tree;	
 
 	NodeInsert(Node,tree,newnode);
 	return newnode;
}		

YggNode *NodeCreate()
{
	YggNode *newnode;
	int len;
 	newnode = (YggNode *) Tcl_Alloc(sizeof(YggNode));
 	
 	/* 
 	 * Fill with nulls
 	 */
 	for(len = 0; len < YGGMAXTOKENLEN;len++) {
 		newnode->name[len] = 0;
 	}
 	newnode->data = NULL;
 	newnode->itree = NULL;
 	newnode->jtree = NULL;
 	newnode->parent = NULL;
 	newnode->checksum = YGGCHECKSUM; /* Used to ensure memory IS a node */
  	newnode->parenttype = 0;	
 	newnode->contents = yggEmpty;
 	YggNodes++;
 	return newnode;
 }

void NodeDelete(thisNode)
	YggNode *thisNode;
{
	YggNode *tempNode,*stack,*parseNode;
	stack = thisNode;
	stack->nextPtr = NULL;
	while(stack != NULL) {
		/*
		 * Step 1 - Thread all branches into a single thread
		 */		
		if(stack->itree != NULL) {
			parseNode = stack->itree;
			while(parseNode->nextPtr != NULL) {			
				parseNode = parseNode->nextPtr;
			}
			parseNode -> nextPtr = stack -> nextPtr;
			stack->nextPtr = stack->itree;
			stack -> itree = NULL;
		}
		if(stack->jtree != NULL) {
			parseNode = stack->jtree;
			while(parseNode->nextPtr != NULL) {
				parseNode = parseNode->nextPtr;
			}
			parseNode -> nextPtr = stack -> nextPtr;
			stack->nextPtr = stack->jtree;
			stack -> jtree = NULL;
		}
		tempNode = stack;
		stack = stack->nextPtr;
		NodeFree(tempNode);
	}
}

void NodeFree(node)
	YggNode *node;
{
		
	YggNodes--;
	if(!NodeInvalid(node)) {
		NodeFreeData(node);		
		Tcl_Free((char *)node);
	}
}

void NodeFreeData(node)
	YggNode *node;
{
	switch	(node->contents) {
		case yggTclObj: {
			Var *varPtr = (Var *)node->data;
			Tcl_DecrRefCount(varPtr->value.objPtr);
			Tcl_Free((char *)varPtr);
		}
		/* 
		 * Storing literal data
		 * No op
		 */
		case yggLink:
		case yggEmpty:
			break;
	}
}

/*
 * Deletes a node from the specified list
 * and returns the new head pointer
 */
YggNode * NodeRemove(pNode,tree,thisNode)
	YggNode *pNode;
	int tree;
	YggNode *thisNode;
{
	YggNode *tempNode,*headPtr;
	
	if(tree == yggItree) {
		headPtr = pNode->itree;
	} else {
		headPtr = pNode->jtree;
	}
	
	if(headPtr == thisNode) {
		headPtr = thisNode->nextPtr;
		thisNode -> nextPtr = NULL;
		/* Handle the simplest case internally */
		if(tree == yggItree) {
			pNode->itree = headPtr;
		} else {
			pNode->jtree = headPtr;
		}
	} else {
		for(tempNode = headPtr;tempNode->nextPtr != NULL;tempNode = tempNode->nextPtr) {
			if(tempNode->nextPtr == thisNode) {
				tempNode->nextPtr = tempNode->nextPtr->nextPtr;
			}
		}
	}
	return headPtr;
}

void NodeInsert(pNode,tree,thisNode)
	YggNode *pNode;
	int tree;
	YggNode *thisNode;
{
	switch(tree) {
		case yggHead:
			Insert(pNode,thisNode);
			break;
		case yggItree:
		 	thisNode->parent 		= pNode;
		  	thisNode->parenttype	= tree;
		  	Insert(pNode->itree,thisNode);
			/*pNode->itree = NodeSort(pNode->itree);*/
			break;
		case yggJtree:
		 	thisNode->parent 		= pNode;
		  	thisNode->parenttype	= tree;
			Insert(pNode->jtree,thisNode);		  	
			/*pNode->jtree = NodeSort(pNode->jtree);*/
			break;
		case yggKtree:
			Insert(pNode->nextPtr,thisNode);
			break;		
	}
}

YggNode *NodeSort(headptr)
	YggNode *headptr;
{
	YggNode *thisPtr,*tempPtr;
	/* Screws me up */
	if(headptr->nextPtr == NULL) {
		return headptr;
	}
	if(strncmp(headptr->name,headptr->nextPtr->name,YGGMAXTOKENLEN) > 0) {
		tempPtr = headptr;
		headptr = headptr -> nextPtr;
		headptr->nextPtr = tempPtr;
	}	
	for(
		thisPtr = headptr->nextPtr;
		thisPtr->nextPtr != NULL;
		thisPtr = thisPtr->nextPtr
	) 
	{
		if(thisPtr == NULL) return headptr;
		if(strncmp(thisPtr->name,thisPtr->nextPtr->name,YGGMAXTOKENLEN) > 0) {	
			tempPtr = thisPtr;
			thisPtr = thisPtr -> nextPtr;
			thisPtr->nextPtr = tempPtr;
		}
	}
	return headptr;
}

int NodeInvalid(node) 
	YggNode *node;
{
	if(node == NULL) {
		return 1;
	}
	if(node->checksum != YGGCHECKSUM) {
		return 1;				
	}
	return 0;
}

/*
 *	Search a branch for a name
 */

YggNode *NodeSearchTree(thisSearch,token,thisNode,err)
	int thisSearch;
	char *token;
	YggNode *thisNode;
	int *err;
{
	YggNode *searchNode,*resultNode;
	int found = 0;
	
	*err = TCL_OK;

	if(thisSearch == yggError) {
		/* Encountered a bad token or address */
		Ygg_Error(yggErrBadToken,token);
		*err = TCL_ERROR;			
		return NULL;
	}	
	if(thisSearch == yggItree) {
		searchNode = thisNode->itree;
	} else {
		searchNode = thisNode->jtree;
	}

	while(searchNode != NULL) {
		if(strncmp(token,searchNode->name,YGGMAXTOKENLEN) == 0) {
			found=1;
			if(searchNode->contents == yggLink) {
				/* Redirect Links */
				resultNode = (YggNode *)searchNode->data;
				if (NodeInvalid(resultNode))  {
					/*
					 * Remove a dead link
					 */
					resultNode->contents = yggEmpty;
					resultNode->data = NULL;
					found=0;				
				}					
			} else {
				resultNode = searchNode;
			}
			break;			
		}
		searchNode = searchNode->nextPtr;
	}

	if(found) {
		return resultNode;
	}
	return NULL;
}


void NodeCopy(srcNode,destNode) 
	YggNode *srcNode;
	YggNode *destNode;
{
	Tcl_Obj *newobj;
	YggNode *thisNode,*newNode;
	
	switch(srcNode->contents) {
		case yggTclObj :
			newobj = Tcl_DuplicateObj(Ygg_GetObj(srcNode));
			Ygg_SetObj(destNode,newobj);
			break;
		case yggLink:
			NodeCopy((YggNode *)srcNode->data,destNode);
			break;
	}
	
	/*
	 * Recursively copy itree
	 */
	if(srcNode->itree != NULL) {
		thisNode = srcNode->itree;
		while(thisNode != NULL) {
			newNode = NodeAdd(destNode,yggItree,thisNode->name,NULL);
			NodeCopy(thisNode,newNode);
			thisNode = thisNode->nextPtr;
		}
	}
	/*
	 * Recursively copy jtree
	 */	
	if(srcNode->jtree != NULL) {
		thisNode = srcNode->jtree;
		while(thisNode != NULL) {
			newNode = NodeAdd(destNode,yggJtree,thisNode->name,NULL);
			NodeCopy(thisNode,newNode);
			thisNode = thisNode->nextPtr;
		}		
	}	
}


Tcl_DString * NodeAddress (record)
	YggNode *record;
{
	YggNode *thisNode;
	YggPath *headptr,*thisptr;
	Tcl_DString *dPtr;
	
	Ygg_NewPath(headptr,(record)->name,(record)->parenttype);
	for(thisNode = record->parent; thisNode != NULL ; thisNode = thisNode->parent) {
		Ygg_NewPath(thisptr,(thisNode)->name,(thisNode)->parenttype);
		Insert(headptr,thisptr);
	}
	
	dPtr = (Tcl_DString *) Tcl_Alloc(sizeof(Tcl_DString));
	Tcl_DStringInit(dPtr);
	Tcl_DStringSetLength(dPtr,0);

	/* Create text buffer */
	/*
	 * We start from root and work down
	 */

	for(thisptr = headptr->nextPtr ; thisptr != NULL ; thisptr = thisptr->nextPtr) {
		switch (thisptr->relation) {
			case yggItree:
				Tcl_DStringAppend(dPtr, "/",-1);	
				break;
			case yggJtree:
				Tcl_DStringAppend(dPtr, ".",-1);
				break;			
		}
		Tcl_DStringAppend(dPtr, thisptr->name,-1);
	}
	/* Add Terminating NULL */
	return dPtr;
}

/*
 * Internal Macro-Like Routines
 */

/*
 * Export Data from tree
 */

void BranchExpand(node,recurse,dPtr) 
	YggNode *node;
	int recurse;
	Tcl_DString *dPtr;
{
	char *val;
	int len;
	if(node == NULL) {
		return;
	}

	Tcl_DStringAppend(dPtr, "\n",-1);	
	Tcl_DStringStartSublist(dPtr);
	if(node->parenttype == yggItree) {
		Tcl_DStringAppendElement(dPtr, "i");
	}
	if(node->parenttype == yggJtree) {
		Tcl_DStringAppendElement(dPtr, "j");
	}
	Tcl_DStringAppendElement(dPtr, node->name);
	
	if(node->contents == yggTclObj) {
		val = Tcl_GetStringFromObj(Ygg_GetObj(node),&len);
		Tcl_DStringAppendElement(dPtr, val);
	}
	Tcl_DStringEndSublist(dPtr);
	
	if( 
		((recurse & yggItree) != 0) && (node->itree != NULL)
		||
		((recurse & yggJtree) != 0) && (node->jtree != NULL)
	) {
		Tcl_DStringAppend(dPtr, "\n<",-1);	
	}
	
	if((recurse & yggItree) != 0) {
		BranchExpand(node->itree,recurse,dPtr);		
	}
	if((recurse & yggJtree) != 0) {
		BranchExpand(node->jtree,recurse,dPtr);		
	}
	if( 
		((recurse & yggItree) != 0) && (node->itree != NULL)
		||
		((recurse & yggJtree) != 0) && (node->jtree != NULL)
	) {
		Tcl_DStringAppend(dPtr, "\n>", -1);	
	}	
	BranchExpand(node->nextPtr,recurse,dPtr);
}

/*
 * Parse through a name string and etch out the next token
 *
 * Made an inline to cut down on name resolution collisions
 * (with multiple tasks there existed the possibility that
 * one lookup would be pre-empted by another stamping 
 * memory like it was going outta style)
 *
 * Compilers that do not support the keywork inline should
 * use a "define" statement to declare it empty.
 * Sorry for the mess... but expediency is expediency
 */
inline int TokenNext(name,token,idx,namelen) 
	char *name;
	char *token;
	int *idx;
	int namelen;
{
	char nchar;
	int ndx=0;
	for(ndx=0;ndx < YGGMAXTOKENLEN;ndx++) {
		(*idx)++;
		if( *idx >= namelen ) {
			token[ndx] = '\0';
			return yggEnd;
		}
		nchar = name[*idx];
		/*
		 * Detect Illegal Characters
		 */
		switch (nchar) {
			case '/':
				token[ndx] = NULL;
				return yggItree;
			case '.':
				token[ndx] = NULL;
				return yggJtree;
			case '\n':
			case '\r':
			case '\t':
			case ' ':
			case '\0':
				token[ndx] = '\0';
				return yggEnd;
			default:
				if(nchar < 32) {
					//sprintf(scriptbuf,"%s %d",token,nchar);
					Ygg_Error(yggIllegalToken,"");
					return yggError;
				}
				if (nchar > 126) {
					//sprintf(scriptbuf,"%s %d",token,nchar);				
					Ygg_Error(yggIllegalToken,"");
					return yggError;
				}
				token[ndx] = nchar;
		}
	}			
	/*
	 * token is too big to be a tree name
	 */
	Ygg_Error(yggErrBadToken,name);	
	return yggError;
}

/*
 * Tcl Variable and command Mapping Routines
 * Unimplimented portions are contained within
 * a precompiler flag NEVER so that they
 * are not built
 */
 
#ifdef NEVER
/* GNDN */
Tcl_Var ResolveRuntimeVarProc (interp,vinfoPtr)
	Tcl_Interp* interp;
	struct Tcl_ResolvedVarInfo *vinfoPtr;
{
	return (Tcl_Var)NULL;
}

/* GNDN */
void ResolveVarDeleteProc (vinfoPtr)
    struct Tcl_ResolvedVarInfo *vinfoPtr;
{
	return;
}

/* GNDN */
int ResolveCmdProc(interp,name,context,flags,rPtr)
	Tcl_Interp* interp;
	char* name;
	Tcl_Namespace *context;
	int flags;
	Tcl_Command *rPtr;
{
	return TCL_CONTINUE;
}

int ResolveCompiledVarProc(interp,name,length,context,rPtr)
    Tcl_Interp* interp;
    char* name;
    int length;
    Tcl_Namespace *context;
    Tcl_ResolvedVarInfo **rPtr;
{
	YggNode *thisNode;
	
	return TCL_CONTINUE;
}
#endif

int ResolveVarProc(interp,name,context,flags,rPtr)
    Tcl_Interp* interp;
    char* name;
    Tcl_Namespace *context;
    int flags;
    Tcl_Var *rPtr;
{
	YggNode *thisNode;
	Tcl_Obj *result;
	Var		*varPtr;
	result = Tcl_GetObjResult(interp);
					
	if (name[0] != '/') {
		return TCL_CONTINUE;
	}
	thisNode = Ygg_Record(name,-1,0);

	if (thisNode == NULL) {	
		goto error;
	}
	
	if (thisNode->contents != yggTclObj) {
		/*
		 * Tried to access a node the was NOT
		 * formatted as a tcl variable
		 */		
		Ygg_Error(yggErrNotTclVar,name);	
		goto error;
	}
	/*
	 * Hmm... Incrementing Reference Count To Lock variable
	 */
	varPtr =(Var *)thisNode->data;
	varPtr->refCount++;
	*rPtr = (Tcl_Var)thisNode->data;
	return TCL_OK;

error:
	Tcl_ResetResult(interp);
	Tcl_SetStringObj(result,YggErrorString, -1);
	Tcl_AppendToObj(result,YggErrorInfo,-1);		
	return TCL_ERROR;
}
